"""
Line profile functions normalized to 1.0 with wavenumber (array) and position, strengths, and width(s) as input variables.
"""

##############################################################################################################
#####  LICENSE issues:                                                                                   #####
#####                 This file is part of the Py4CAtS package.                                          #####
#####                 Copyright 2002 - 2019; Franz Schreier;  DLR-IMF Oberpfaffenhofen                   #####
#####                 Py4CAtS is distributed under the terms of the GNU General Public License;          #####
#####                 see the file ../license.txt in the parent directory.                               #####
##############################################################################################################

# some math constants
from math import pi, sqrt
from ir import h, c, k, sqrtPi, ln2, sqrtLn2, recSqrtPi
from racef import hum1wei24, hum2wei32, weideman32a

recSqrt2      = 1./sqrt(2.0)
recSqrtLn2    = 1./sqrtLn2
sqrtLn2overPi = sqrtLn2/sqrtPi

try:                      import numpy as np
except ImportError as msg:  raise SystemExit (str(msg) + '\nimport numpy (numeric python) failed!')


####################################################################################################################################

def voigtWidth (gammaL=1.0, gammaD=1.0):
	""" Half width half maximum (HWHM) of Voigt profile (Whiting's approximation). """
	return 0.5*(gammaL+np.sqrt(gammaL**2+4.0*gammaD**2))


####################################################################################################################################

def Lorentz (vGrid, vLine=0.0, S=1.0, gammaL=1.0):
	""" Pressure broadening: Lorentz profile normalized to one, multiplied with line strength. """
	return (S*gammaL/pi)/((vGrid-vLine)**2+gammaL**2)


def LorentzMix (vGrid, vLine=0.0, S=1.0, gammaL=1.0, yMix=0.0):
	""" Pressure broadening: Lorentz profile incl. line mixing, normalized to one, multiplied with line strength.
	    See D. Edwards, SPIE Vol. 928 (1986), Eq. (4.5) """
	deltaV = vGrid-vLine
	return (S/pi)*(gammaL + yMix*deltaV)/(deltaV**2+gammaL**2)


####################################################################################################################################

def vanVleckWeisskopf (vGrid, vLine=0.0, S=1.0, gammaL=1.0):
	""" Pressure broadening: VanVleck-Weisskopf lineshape without tanh prefactor, multiplied with line strength.
	    ARTS manual, eq. (3.12)
	    ARTS --- Buehler, Eriksson et al., JQSRT Vol. 91 (2005), Eq. (7)
	    See D. Edwards, SPIE Vol. 928 (1986), Eq. (4.3)
	"""
	factor = S * gammaL/pi
	return (vGrid/vLine)**2 * (factor / ((vGrid-vLine)**2+gammaL**2) + factor / ((vGrid+vLine)**2+gammaL**2))


####################################################################################################################################

def vanVleckHuber (vGrid, vLine=0.0, S=1.0, gammaL=1.0, temp=250.0):
	""" Pressure broadening: VanVleck-Huber lineshape with tanh prefactor, multiplied with line strength.
	    See D. Edwards, SPIE Vol. 928 (1986), Eq. (4.3)
	    ARTS --- Buehler, Eriksson et al., JQSRT Vol. 91 (2005), Table 2
	"""
	if not isinstance(temp,float): raise SystemExit('ERROR --- vanVleckHuber: no temperature specified!')
	factor = S * gammaL/pi
	hc2k   = h*c/(2.0*k)
	tv0    = np.tanh(hc2k*vLine/temp)
	return vGrid * np.tanh(hc2k*vGrid/temp)/(vLine*tv0) * \
	       (factor / ((vGrid-vLine)**2+gammaL**2) + factor / ((vGrid+vLine)**2+gammaL**2))


####################################################################################################################################

def Gauss (vGrid, vLine=0.0, S=1.0, gammaD=1.0):
	""" Doppler broadening:  Gauss profile normalized to one, multiplied with line strength. """
	t = ln2*((vGrid-vLine)/gammaD)**2
	# don't compute exponential for very large numbers > log(1e308)=709.
	# (in contrast to python.math.exp the numpy exp overflows!!!)
	return S * sqrtLn2/(sqrtPi*gammaD) * np.where(t<200.,np.exp(-t),0.0)


####################################################################################################################################

def Voigt (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0):
	""" Voigt profile normalized to one, multiplied with line strength. """
	rgD = sqrtLn2 / gammaD
	x   = rgD * (vGrid-vLine)
	y   = rgD * gammaL
	vgt = recSqrtPi*rgD*S * hum1wei24(x,y).real  # Voigt profile
	return vgt


def VoigtMix (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0, yMix=0.0):
	""" Voigt profile normalized to one, multiplied with line strength. """
	rgD = sqrtLn2 / gammaD
	x   = rgD * (vGrid-vLine)
	y   = rgD * gammaL
	cef = hum1wei24(x,y)
	return recSqrtPi*rgD*S * (cef.real+yMix*cef.imag)


####################################################################################################################################

def speedVoigt (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0, gamma2=1.0):
	""" Speed-dependent profile normalized to one, multiplied with line strength.

	    Boone, Walker, Bernath (JQSRT 112, 2011):
	    An efficient analytical approach for calculating line mixing in atmospheric remote sensing applications

	    Schreier (JQSRT 187, 2017): Computational Aspects of Speed-Dependent Voigt Profiles
	"""

	# sdv effective parameters (Boone et al. Eq. 13)
	alfa = gammaL/gamma2 - 1.5
	beta = (vGrid - vLine)/gamma2
	sqrtDelta = 0.5*recSqrtLn2 * (gammaD/gamma2)
	delta = sqrtDelta**2
	#print ('speedVoigt --- alfa, delta: ', alfa, delta)

	# arguments for the complex error function
	r4  = np.sqrt((delta+alfa)**2+beta**2)
	izz = recSqrt2 * (np.sign(beta)*np.sqrt(np.maximum(r4-delta-alfa,0.0)) +1j*np.sqrt(r4+delta+alfa))
	izp = izz + 1j*sqrtDelta
	izm = izz - 1j*sqrtDelta

	# evaluate the difference terms
	wPlus  = weideman32a(izp.real, izp.imag)
	wMinus = weideman32a(izm.real, izm.imag)
	# compute difference of real parts and normalize
	profile = sqrtLn2*recSqrtPi*(wMinus-wPlus).real/gammaD

	return S*profile


def speedVoigtMix (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0, gamma2=1.0, yMix=0.0):
	""" Speed-dependent profile including Rosenkranz line mixing normalized to one, multiplied with line strength.

	    Boone, Walker, Bernath (JQSRT 112, 2011):
	    An efficient analytical approach for calculating line mixing in atmospheric remote sensing applications

	    F. Schreier (JQSRT 187, 2017):  Computational aspects of speed-dependent Voigt profiles
	"""

	# sdv effective parameters (Boone et al. Eq. 13)
	alfa = gammaL/gamma2 - 1.5
	beta = (vGrid - vLine)/gamma2
	sqrtDelta = 0.5*recSqrtLn2 * (gammaD/gamma2)
	delta = sqrtDelta**2
	#print ('speedVoigtMix --- alfa, delta: ', alfa, delta)

	# arguments for the complex error function
	r4  = np.sqrt((delta+alfa)**2+beta**2)
	izz = recSqrt2 * (np.sign(beta)*np.sqrt(np.maximum(r4-delta-alfa,0.0)) +1j*np.sqrt(r4+delta+alfa))
	izp = izz + 1j*sqrtDelta
	izm = izz - 1j*sqrtDelta

	# evaluate the difference terms
	wPlus  = weideman32a(izp.real, izp.imag)
	wMinus = weideman32a(izm.real, izm.imag)
	# compute difference of real parts and multiply with mixing factor (Boone et al. Eq. 10)
	profile = (1 - 1j*yMix) * (wMinus-wPlus)

	return S * (sqrtLn2*recSqrtPi/gammaD) * profile.real


####################################################################################################################################

def Rautian (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0, gammaN=1.0):
	""" Rautian profile normalized to one, multiplied with line strength.
	    Broadening by thermal motion and state-perturbing collisions;
	    hard collision model for velocity-changing collisions; collisional perturbations uncorrelated

	    Philip L. Varghese and Ronald K. Hanson:
	    Collisional narrowing effects on spectral line shapes measured at high resolution [AO 23(14), 2376-2385, 1984]
	"""
	rgD   = sqrtLn2 / gammaD
	x     = rgD * (vGrid-vLine)
	y     = rgD * gammaL
	zeta  = rgD * gammaN
	cef   = hum1wei24(x,y+zeta)
	ratio = cef / (1.0-sqrtPi*zeta*cef)
	return recSqrtPi*rgD*S * ratio.real


####################################################################################################################################

def Voigt_Kuntz_Humlicek1 (vGrid, vLine=0.0, S=1.0, gammaL=1.0, gammaD=1.0):
	""" Voigt profile normalized to one, multiplied with line strength.
	    Rational approximation for asymptotic region for large |x|+y.
	    Real part only using Kuntz (JQSRT, 1997) implementation with Ruyten's (JQSRT, 2003) correction. """
	xx = (sqrtLn2 * (vGrid-vLine)/ gammaD)**2
	y  = sqrtLn2 * gammaL / gammaD
	yy = y*y
	kha1 = 0.5+yy
	kha2 = kha1 + yy*yy; khb2 = 2.*yy-1.0
	return  (sqrtLn2overPi/gammaD)*S * y*recSqrtPi * (kha1+xx) / (kha2 + xx*(khb2+xx))
