"""  radInt

  functions to read, write or plot radiation intensity (e.g., to reformat, truncate, or interpolate).
"""

_LICENSE_ = """\n
This file is part of the Py4CAtS package.

Authors:
Franz Schreier
DLR-IMF Oberpfaffenhofen
Copyright 2002 - 2019  The Py4CAtS authors

Py4CAtS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Py4CAtS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

####################################################################################################################################

# import some standard python modules
import os, sys
from math import ceil
from string import punctuation

try:   import numpy as np
except ImportError as msg:  raise SystemExit(str(msg) + '\nimport numeric python failed!')

try:
	from matplotlib.pyplot import plot, legend, xlabel, ylabel
except ImportError:
	print ('WARNING --- radInt:  matplotlib not available, no quicklook!')
	#raise SystemExit ('ERROR --- radInt:  matplotlib not available, no quicklook!')


from aeiou import awrite, cstack, grep_from_header, join_words, read_first_line
from pairTypes import Interval
from cgsUnits import cgs
from grid import is_uniform
from misc import regrid
from convolution     import convolveBox, convolveTriangle, convolveGauss
from radiance2radiance import radiance2radiance
from radiance2Kelvin import radiance2Kelvin

####################################################################################################################################
####################################################################################################################################

class riArray (np.ndarray):
	""" A subclassed numpy array of radiance intensity with xLimits, z, ... attributes added.

	    Furthermore, some convenience functions are implemented:
	    *  dx:       return wavenumber grid point spacing
	    *  grid:     return a numpy array with the uniform wavenumber grid
	    *  regrid:   return a riArray with the ri data interpolated to a new grid (same xLimits!)
	    #  truncate: return a riArray with the wavenumber range (xLimits) truncated
	    #  convolve: return a riArray of radiance convolved with a spectral response function
	"""
	# http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

	def __new__(cls, input_array, xLimits=None, tBack=None, zenith=None, z=None, p=None, t=None, srf=""):
		# Input array is an already formed ndarray instance
		# First cast to be our class type
		obj = np.asarray(input_array).view(cls)
		# add the new attributes to the created instance
		obj.x      = xLimits
		obj.tBack  = tBack
		obj.zenith = zenith
		obj.z      = z
		obj.p      = p
		obj.t      = t   # temperature (cannot use capital "T" because this means 'transpose')
		obj.srf    = srf # spectral response function
		return obj  # Finally, we must return the newly created object:

	def __array_finalize__(self, obj):
		# see InfoArray.__array_finalize__ for comments
		if obj is None: return
		self.x      = getattr(obj, 'x', None)
		self.tBack  = getattr(obj, 'tBack', None)
		self.zenith = getattr(obj, 'zenith', None)
		self.z      = getattr(obj, 'z', None)
		self.p      = getattr(obj, 'p', None)
		self.t      = getattr(obj, 't', None)
		self.srf    = getattr(obj, 'srf', "")

	def __str__ (self):
		return '# %i grid points for x=%s  (backTemp=%.1fK,  zenith=%sdg)\n %s' % \
		       (len(self), self.x, self.tBack, self.zenith, self.__repr__())
		#return '# (z=%s,   p=%s,   zenith=%s)  %i \n %s' % (self.z, self.p, self.zenith, len(self), self.__repr__())

	def info (self):
		""" Return basic information (z, p, T, wavenumber and radiance range). """
		return '%6.1f -- %.1fkm %12.3e -- %9.3emb %8.1f -- %5.1fK  %12i points in  %10f -- %10f cm-1  with  %10.4g < ri < %9.4g  %s' % \
                       (cgs('!km',self.z.lower), cgs('!km',self.z.upper),
		        cgs('!mb',self.p.lower), cgs('!mb',self.p.upper),  self.t.lower, self.t.upper,
		        len(self), self.x.lower, self.x.upper, min(self), max(self), self.srf)

	def dx (self):
		""" Return wavenumber grid point spacing. """
		return  self.x.size()/(len(self)-1)

	def grid (self):
		""" Setup a uniform, equidistant wavenumber grid of len(self). """
		return  self.x.grid(len(self))  # calls the grid method of the Interval class (in pairTypes.py)

	def regrid (self, new, method='l', yOnly=False):
		""" Interpolate radiance (intensity) to (usually denser) uniform, equidistant wavenumber grid. """
		yNew = regrid(self.base,new,method)
		if yOnly:  return  yNew
		else:      return  riArray (yNew, self.x, self.tBack, self.zenith, self.z, self.p, self.t, self.srf)

	def truncate (self, xLimits):
		""" Return a radiance / intensity in a truncated (smaller) wavenumber interval. """
		if isinstance(xLimits,(tuple,list)) and len(xLimits)==2:  xLimits=Interval(*xLimits)
		dx = self.dx()
		iLow  = max(int((xLimits.lower-self.x.lower)/dx), 0)
		iHigh = min(int(ceil((xLimits.upper-self.x.lower)/dx)), len(self)-1)
		xLow  = self.x.lower + iLow*dx
		xHigh = self.x.lower + iHigh*dx
		return  riArray (self.base[iLow:iHigh+1], Interval(xLow,xHigh),
		                 self.tBack, self.zenith, self.z, self.p, self.t, self.srf)

	def convolve (self, hwhm=1.0, srf='Box'):
		""" Return a radiance / intensity convolved with a spectral response function of half width @ half maximum. """
		if self.x.size()<10*hwhm:  raise SystemExit ("ERROR --- riArray.convolve:  hwhm too large!")
		if srf.upper().startswith('G'):
			srfInfo = '%s %s' % ('Gauss:  ', hwhm)
			wGrid, rad = convolveGauss (self.grid(), self.base, hwhm)
		elif srf.upper().startswith('T'):
			srfInfo = '%s %s' % ('Triangle:  ', hwhm)
			wGrid, rad  =  convolveTriangle(self.grid(), self.base, hwhm)
		else:
			srfInfo = '%s %s' % ('Box:  ', hwhm)
			wGrid, rad  =  convolveBox(self.grid(), self.base, hwhm)
		return  riArray (rad, Interval(wGrid[0],wGrid[-1]), self.tBack, self.zenith, self.z, self.p, self.t, srfInfo)

	def __eq__(self, other):
		""" Compare radiance/intensity including its attributes.
		    (For p and od relative differences < 0.1% are seen as 'equal') """
		return self.x==other.x \
			   and self.z.lower==other.z.lower and self.z.upper==other.z.upper \
			   and self.t.lower==other.t.lower and self.t.upper==other.t.upper \
			   and abs(self.p.lower-other.p.lower)<0.001*self.p.lower \
			   and abs(self.p.upper-other.p.upper)<0.001*self.p.upper \
			   and abs(self.tBack-other.tBack)<0.1 \
			   and abs(self.zenith-other.zenith)<0.1 \
			   and self.srf==other.srf \
			   and np.allclose(self.base,other.base,atol=0.0,rtol=0.001)


####################################################################################################################################

def riPlot (radi, mue=False, kelvin=False):
	""" Quicklook plot radiance intensity vs. wavenumber.

 	    ARGUMENTS:
 	    ----------

	    radi:          radiance intensity spectrum (along with attributes) or a list thereof
	    mue:           flag:  plot radiance vs. wavelength [mue] (default False)
	    kelvin:        flag:  plot equ. brightness temperature vs. wavenumber (default False)
	"""
	if isinstance(radi,(list,tuple)):
		# recursively call the plot function
		for ri in radi:
			riPlot (ri, mue, kelvin)
	elif isinstance(radi,riArray):
		if isinstance(radi.zenith,(int,float)) and isinstance(radi.tBack,(int,float)):
			labelText = '%.1fdg %.1fK  %s' % (radi.zenith, radi.tBack, radi.srf)
		else:
			labelText = radi.srf
		if mue:
			if kelvin:  print("WARNING --- riPlot:  brightness temperature vs. wavelength not implemented")
			lambdaGrid, radiance = radiance2radiance(radi.grid(), radi.base, newX='mue')
			plot (lambdaGrid, radiance, label=labelText)
			xlabel (r'wavelength ~ $\lambda \rm\,[\mu m]$')
			ylabel (r'radiance ~ $I(\lambda) \rm\, [erg/s/(cm^2\:sr\:\mu m)]$')
		else:
			if kelvin:
				vGrid = radi.grid()
				plot (vGrid, radiance2Kelvin(vGrid, radi.base), label=labelText)
				ylabel (r'Equ. Temperature ~ $T_{\rm B}(\nu) \rm\, [erg/s/(cm^2\:sr\:cm^{-1})]$')
			else:
				plot (radi.grid(), radi.base, label=labelText)
				ylabel (r'radiance ~ $I(\nu) \rm ~ [erg/s/(cm^2\:sr\:cm^{-1})]$')
			xlabel (r'wavenumber ~ $\nu \rm\,[cm^{-1}]$')
	else:
		raise SystemExit ("ERROR --- od2ri.riPlot:  expected a riArray or a list thereof!")
	legend (fontsize='small')


####################################################################################################################################
####################################################################################################################################

def riRead (riFile, commentChar='#'):
	""" Read radiance intensity vs. wavenumber along with some attributes from file. """

	if not os.path.isfile(riFile):  raise SystemExit ('ERROR --- riRead:  riFile "' + riFile + '"not found')

	# try to determine filetype automatically from first nonblank character in file
	firstLine = read_first_line (riFile)

	if firstLine.startswith(commentChar):
		radiance = _riRead_xy (riFile, commentChar)
	else:
		radiance = _riRead_pickled (riFile)

	return  radiance


####################################################################################################################################

def _riRead_pickled (riFile):
	""" Read radiance intensity vs. wavenumber along with some attributes from pickle file. """

	try:                    pf = open(riFile,'rb')
	except IOError as msg:  raise SystemExit ('%s %s %s\n %s' % ('ERROR:  opening pickled radiance file ',
	                                                           repr(riFile), ' failed (check existance!?!)', msg))
	from pickle import load

	# NOTE:  the sequence of loads has to corrrespond to the sequence of dumps!
	info  = load(pf); print(riFile, info)

	# initialize list of radiances to be returned
	riList = []

	while 1:
		try:
			riDict = load(pf)
			print (type(riDict), riDict)
			for key,val in riDict.items():  print (key, val)
			riList.append(riArray(riDict['y'], riDict['x'], riDict['tBack'], riDict['angle'],
			                      riDict['altitude'], riDict['pressure'], riDict['temperature']))
		except EOFError:
			pf.close(); print('INFO --- riRead_pickled:  EOF, got', len(riList), 'radiance(s)'); break
	return riList


####################################################################################################################################

def _riRead_xy (riFile, commentChar='#'):
	""" Read radiance intensity vs. wavenumber along with some attributes from ascii tabular file.

	    NOTE:  multi-column data (i.e. several radiances vs. wavenumber) are not yet implemented!
	"""

	#xyData, riComments = readDataAndComments(riFile)
	vGrid, riValues = np.loadtxt(riFile, usecols=(0,1), unpack=1, comments=commentChar)

	if is_uniform(vGrid,0.002, 1):  vLimits = Interval(vGrid[0],vGrid[-1])
	else:                           raise SystemExit ("ERROR --- riRead_xy:  wavenumber grid is not equidistant/uniform")

	angle     = grep_from_header(riFile,'angle')[0][0]
	backTemp  = grep_from_header(riFile,'back temperature')[0][0]
	try:
		zInfo     = grep_from_header(riFile,'zInterval');  zInterval = Interval(*cgs(zInfo[1],zInfo[0]))
	except Exception as msg:
		zInterval = Interval(0,0);  print (str(msg) + "\nWARNING --- riRead:  failed to read zInterval")
	try:
		pInfo     = grep_from_header(riFile,'pInterval');  pInterval = Interval(*cgs(pInfo[1],pInfo[0]))
	except Exception as msg:
		pInterval = Interval(0,0);  print (str(msg) + "\nWARNING --- riRead:  failed to read pInterval")
	try:
		tInfo     = grep_from_header(riFile,'tInterval');  tInterval = Interval(*tInfo[0])
	except Exception as msg:
		tInterval = Interval(0,0);  print (str(msg) + "\nWARNING --- riRead:  failed to read tInterval")
	srfInfo = grep_from_header(riFile,'specResponseFct')
	if not srfInfo:  srfInfo=''

	return  riArray (riValues, vLimits, backTemp, angle, zInterval, pInterval, tInterval,srfInfo)


####################################################################################################################################
####################################################################################################################################

def riSave (radi, outFile=None, info=None, commentChar='#'):
	""" Write radiance intensity vs. wavenumber along with some attributes to file.

 	    ARGUMENTS:
 	    ----------

	    radi:          radiance intensity spectrum (along with attributes) or a list thereof
	    outFile:       destination file, stdout if unspecified
	    info:          information string to be included in file header (default None)
	    commentChar    character to be used for comment lines in file header
	                   default empty string "", i.e. pickle radiance(s)
	"""

	if isinstance(commentChar,str) and len(commentChar)>0 and commentChar[0] in punctuation:
		_riSave_xy (radi, outFile, info, commentChar)
	else:
		_riSave_pickled (radi, outFile, info)


####################################################################################################################################

def _riSave_xy (radi, outFile=None, info=None, commentChar='#'):
	""" Write radiance intensity vs. wavenumber along with some attributes to ascii file.  """

	# atmospheric data for the file header
	if isinstance (info, str):  comments = [info]
	else:                       comments = []

	comments.append('wavenumber [cm-1]')
	comments.append('radiance [erg/s/(cm2 sr cm-1)]')

	if isinstance(radi, (list,tuple)) and all([isinstance(ri,riArray) for ri in radi]):
		# check if all wavenumber grids are identical
		if all([rad.x==radi[0].x and len(rad)==len(radi[0]) for rad in radi[1:]]):
			comments.insert(-2,'angle [dg]:     ' + len(radi)*' %10.1f ' % tuple([rad.zenith for rad in radi]))
			comments.insert(-2,'back temperature [K]:' + len(radi)*' %10.1f ' % tuple([rad.tBack for rad in radi]))
			awrite ((radi[0].grid(),cstack(radi)), outFile,
			        comments=comments, format='%10f %10.5g', commentChar=commentChar)
		else:
			for rad in radi:  print(rad.x, len(rad))
			raise SystemExit ("ERROR --- od2ri.riSave:  radiance spectra do have different wavenumber grids!")
	elif isinstance(radi,riArray):
		# save further attributes if given
		if isinstance (radi.tBack, (int,float)):   comments.insert(-2,'back temperature [K]: %10.2f' % radi.tBack)
		if isinstance (radi.zenith, (int,float)):  comments.insert(-2,'angle [dg]:     %10.2f' % radi.zenith)
		if isinstance (radi.z, Interval):
			comments.insert(-2,'zInterval [cm]: %10.2f %10.2f' % (cgs('!km',radi.z.lower),cgs('!km',radi.z.upper)))
		if isinstance (radi.p, Interval):
			comments.insert(-2,'pInterval [mb]: %10.2g %10.2g' % (cgs('!mb',radi.p.lower),cgs('!mb',radi.p.upper)))
		if isinstance (radi.t, Interval):  comments.insert(-2,'tInterval  [K]: %10.2f %10.2f' % (radi.t.lower,radi.t.upper))
		if isinstance (radi.srf, str) and len(radi.srf)>0:  comments.insert(-2,'specResponseFct:  %s' % radi.srf)
		# now write ascii tabular file
		awrite ((radi.grid(),radi), outFile, comments=comments, format='%10f %10.5g', commentChar=commentChar)
	else:
		raise SystemExit ("ERROR --- od2ri.riSave:  expected a riArray or a list thereof!")


####################################################################################################################################

def _riSave_pickled (radi, outFile, info=None):
	""" Write radiance / intensity spectrum to file using pickle format. """
	if outFile:  out = open (outFile, 'wb')
	else:        raise SystemExit ('\n ERROR --- riSave:  no radiance/intensity pickling for standard out!')

	import pickle
	if 'ipython' in join_words(sys.argv) or 'ipykernel' in join_words(sys.argv):
		if isinstance (info, str):  pickle.dump('ipy4cats:  ' + info, out)
		else:                       pickle.dump('ipy4cats', out)
	else:
		pickle.dump(join_words([os.path.basename(sys.argv[0])] + sys.argv[1:]), out)

	if isinstance(radi, (list,tuple)) and all([isinstance(ri,riArray) for ri in radi]):
		for ri in radi:
			pickle.dump ({'angle': ri.zenith, 'tBack': ri.tBack, 'altitude': ri.z,
			              'pressure': ri.p, 'temperature': ri.t, 'x': ri.x,  'y': ri.base}, out)
	elif isinstance(radi,riArray):
		pickle.dump ({'angle': radi.zenith, 'tBack': radi.tBack, 'altitude': radi.z,
		              'pressure': radi.p, 'temperature': radi.t, 'x': radi.x,  'y': radi.base}, out)
	else:
		raise SystemExit ("ERROR --- od2ri.riSave:  expected a riArray or a list thereof!")
	out.close()
